require("AntPathFinder.nut");
require("CrowPathFinder.nut");
require("BridgePathFinder.nut");

class AntHill
{
  paths = null;
  ant = null;
  crow = null;
  start = null;
  goal = null;
  cost = null;
  last = null;
  ignoreTiles = null;
  
  prepend = null;
  postpend = null;
  
  display_thinking = null;
  
  routesBuilt = 0;
  
  constructor(start, goal, prepend, postpend, ignoreTiles)
  {
    this.start = start;
    this.goal = goal;
    this.paths = [];
    this.ant = AntPathFinder(start, goal, ignoreTiles);
    this.crow = CrowPathFinder(start, goal, ignoreTiles);
    this.cost = null;
    routesBuilt = 0;
    this.prepend = prepend;
    this.postpend = postpend;
    this.display_thinking = AIController.GetSetting("display_thinking");
    
    local startList = AIList();
    startList.AddList(prepend);
    startList.Sort(AIAbstractList.SORT_BY_VALUE, false);
    this.last = startList.Begin();
    crow.last = last;
    ant.last = last;
    this.ignoreTiles = ignoreTiles;
  }

}

function AntHill::FindPathMixed()
{
  print("Hill Last == " + last);
  if(display_thinking == 1)
  {
    AISign.BuildSign(start, "{+}");
    AISign.BuildSign(goal, "{=}");
  }
  local run = true;
  local currentStart = start;
  local currentGoal = goal;
  local pathSegments = [];
  local iteration = 0;
  local iterationLimit = 25;
  
  pathSegments.push(prepend);
  
  while(run && iteration < iterationLimit)
  {
    iteration++;
    crow = CrowPathFinder(currentStart, currentGoal, ignoreTiles);
    print ("starting crow.");
    if(crow.FindPath())
    {
      run = false;
      pathSegments.push(crow.path);
      print ("crow found goal.");
    }
    else
    {
      pathSegments.push(crow.path);
      currentStart = crow.midStart;
      currentGoal = crow.midGoal;
      last = crow.last;
      print("AntHill Last == " + last);
      
      if(currentGoal == null || currentStart == null)
      {
        pathSegments = null;
        run = false;
      }
      else
      {
  
        print("starting bridge.");
        local bridge = BridgePathFinder(currentStart, currentGoal);
        if(bridge.FindPath())
        {
          pathSegments.push(bridge.path);
          currentStart = bridge.goal;
          currentGoal = this.goal;
          print ("Bridge found the way over.");
        }
        else
        {
          print ("Bridge failed.");      
          print ("starting Ant.");
          
          //local town = AITile.GetClosestTown(currentStart);
          //local townManager = TownManager(town);
          //townManager.BoxInTown();
          
          local hill = AntHill(currentStart, currentGoal, AIList(), AIList(), ignoreTiles);
          hill.last = last;
          local path = hill.FindPath(5);
          if(path != null)
          {
            pathSegments.push(path);
            currentStart = currentGoal;
            currentGoal = this.goal;
            print ("ant found way around.");
          }
          else
          {
            print("Ant failed to find a way around.");
            run = false;
            pathSegments = null;
          }
        }
      }
    }
  }
  
  

  if(iteration == iterationLimit) 
  {
    AILog.Warning("Iteration Limit reached.");
    pathSegments = null;
  }
  
  local fullPath = AIList();
  local fullCount = 0;
  
  
  
  if(pathSegments != null)
  {
    pathSegments.push(postpend);
    foreach(i, val in pathSegments)
    { 
      if(val == null)
      {
       continue;
      }
      val.Sort(AIAbstractList.SORT_BY_VALUE, true);
      foreach(tile, x in val)
      {
        fullPath.AddItem(tile, fullCount++);
      }
    }
  
    
    this.cost = AntHill.CostPath(fullPath);
    fullPath.Sort(AIAbstractList.SORT_BY_VALUE, true);
    return fullPath;
  }
  else
  {
    return null;
  }
}


function AntHill::FindPath(iterations)
{
 //print(this.start + "  " + this.goal);
 print("Anthill::FindPath last == " + last);
 for(local i = 0; i < iterations; i++)
 {
  print ("Ant pathfinding Iteration:" + i);
  ant.last = last;
  local path = ant.FindPath();
  paths.push({ path=path, cost = ant.cost });
  ant = AntPathFinder(this.start, this.goal, ignoreTiles);
  //Make each run more goal focused to try more direct routes.
  ant.goalFocus += (i / iterations) * 4.0 ;
 }

 local lowCost = 999999999;
 local bestPath = null;
 foreach(i, val in paths)
 {
    //print("i:" + i + " cost:" + val.cost);
    if(val.cost < lowCost)
    {
      bestPath = val.path;
      lowCost = val.cost;
    }
 }

 this.cost = lowCost;
  
 if(bestPath != null && display_thinking == 1)
 //if(bestPath != null)
 {
   foreach(tile, v in bestPath)
   {
    AISign.BuildSign(tile, "{A}");
   }
 }
 
 return bestPath;
}

function AntHill::VerifyPath(path)
{
    if(path.Count() > 3) {return 1;}
    local test = AITestMode();
    local pp = null;
    local p = null;
    local count = 0;
    foreach(tile, val in path)
    {
      if(p != null && pp != null)
      {
       if (!AIRail.BuildRail(pp, p, tile))
       {
        return false;
       }
      }
      pp = p; p = tile;
    }
    return true;
}

function AntHill::CostPath(path)
{
    
    if(path.Count() < 3) {return path.Count() * 500;}
    local test = AITestMode();
    local cost = AIAccounting();
    local pp = null;
    local p = null;
    local count = 0;
    foreach(tile, val in path)
    {
      if(p != null && pp != null)
      {
       AIRail.BuildRail(pp, p, tile);
      }
      pp = p; p = tile;
    }
    //print ("AntHill: Route Cost:" + costs.GetCosts());
    return cost.GetCosts();
}


function AntHill::BuildRoute(path)
{   
    local pp = null;
    local p = null;
    local count = 0;
    foreach(tile, val in path)
    {
    
    
      if(p != null && pp != null)
      {
         //AISign.BuildSign(p, count++ + "");
         if (AIMap.DistanceManhattan(p, tile) > 1)
         {
            local bridge_list = AIBridgeList_Length(AIMap.DistanceManhattan(tile, p) + 1);
            bridge_list.Valuate(AIBridge.GetMaxSpeed);
            bridge_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
            local result = AIBridge.BuildBridge(AIVehicle.VT_RAIL, bridge_list.Begin(), p, tile);
            if(result == false)
            {
             AILog.Warning("Bridge Failed! Marking path.");
             foreach(tile, val in path)
             {
              AISign.BuildSign(tile, count++ + "");
             }
            }
         }
         else
         {
           
           AIRail.BuildRail(pp, p, tile);
         }
      }
      pp = p; p = tile;
    }
}

function AntHill::FindCompassPointsNunitsAway(point, n)
{
    //print("point =" + point);
    local tiles = AITileList();
    
    local x = AIMap.GetTileX(point);
    local y = AIMap.GetTileY(point);
    
    
    tiles.AddTile(AIMap.GetTileIndex(x , y - n));
    tiles.AddTile(AIMap.GetTileIndex(x - n, y));  
    tiles.AddTile(AIMap.GetTileIndex(x + n, y));
    tiles.AddTile(AIMap.GetTileIndex(x, y + n));
    
    return tiles;
}