//An "as the crow flies" pathfinder for gross pathfinding.

require("AntPathFinder.nut");

class CrowPathFinder
{
  start = null;
  goal = null;
  current = null;
  midStart = null;
  midGoal = null;
  path = null;
  count = null;
  last = null;
  display_thinking = null;
  ignoreTiles = null;
  
  

  constructor(start, goal, ignoreTiles)
  {
    this.start = start;
    this.current = start;
    this.goal = goal;
    this.midGoal = null;
    this.midStart = null;
    this.last = null;
  
    this.count = 0;
    this.path = AIList();
    this.path.AddItem(start, count);
    this.display_thinking = AIController.GetSetting("display_thinking");
    this.ignoreTiles = ignoreTiles;
  }

}

function CrowPathFinder::FindPath()
{
  if(start == goal) {return true;}

  local distance = AITile.GetDistanceSquareToTile(this.current, goal);
  print("Distance to goal is:" + distance);

  if(!AntPathFinder.isBuildableOrMyRail(goal))
  {
    AILog.Warning("Destination is not valid.");
    return false;
  }

  while(CrowPathFinder.MoveToNext())
  {

    local distance = AITile.GetDistanceSquareToTile(this.current, this.goal);
    //if(this.current == this.goal)
    if(distance < 75) //crow isn't great at the "last mile";
    {
      count++;
      this.path.AddItem(this.current, count);
      this.path.Sort(AIAbstractList.SORT_BY_VALUE, false);
      this.midStart = path.Begin();
      this.last = path.Next();
      print("CrowLast == " + this.last);
      this.midGoal = goal;
      return false;
    }
  }
  
  
  //print("crow: ran into something.");

  if(count > 3)
  {
    //trim off some tiles to back off from obstacles.
    count -= 1;
    this.path.KeepBelowValue(count);
  }
  this.path.Sort(AIAbstractList.SORT_BY_VALUE, false);
  
  this.midStart = path.Begin();
  this.last = path.Next();

  this.midGoal = CrowPathFinder.FindNextSafeTile();
  return false;
}

function CrowPathFinder::MoveToNext()
{
  local tile = CrowPathFinder.GetNextClosestTile(this.current);
  
  if(tile != null && AntPathFinder.isBuildable(tile, ignoreTiles) )
  {
    if(display_thinking == 1) { AISign.BuildSign(tile, "{.}")};
    last = current;
    current = tile;
    count++;
    this.path.AddItem(current, count);
    //print("last:"+ last + " current:" +current);
  }
  else
  { 
    return false;
  }
  return true;
}

function CrowPathFinder::DetermineDirection(tile, last)
{
    local cardinal = true;
    if(last == null) {return;}
    local x1 = AIMap.GetTileX(tile);
    local y1 = AIMap.GetTileY(tile);
  
    local x2 = AIMap.GetTileX(last);  
    local y2 = AIMap.GetTileY(last); 
    
    local dx = x1 - x2;
    local dy = y1 - y2;
    
    //print("("+x1+","+y1+"),("+x2+","+y2+")");
    //AISign.BuildSign(tile, dx + "," + dy);
    
    //if(dx == 0 && dy > 0){AISign.BuildSign(tile, "e");}    
    //if(dx == 0 && dy < 0){AISign.BuildSign(tile, "w");}
    //if(dx > 0 && dy == 0){AISign.BuildSign(tile, "s");}
    //if(dx < 0 && dy == 0){AISign.BuildSign(tile, "n");}
    if(dx > 0 && dy > 0){cardinal = false;}
    if(dx > 0 && dy < 0){cardinal = false;}
    if(dx < 0 && dy > 0){cardinal = false;}
    if(dx < 0 && dy < 0){cardinal = false;}
    
    return cardinal;

}


function CrowPathFinder::FindNextSafeTile()
{
  local distanceToGoal = AITile.GetDistanceSquareToTile(this.current, goal);
  print("distance left:" + distanceToGoal);
  if(distanceToGoal < 15) {return goal;}
  
  local safeTiles = 0;
  // local tile = CrowPathFinder.GetNextClosestTile(this.current);
  local offset = current - last;
  local tile = current + offset;
  
  while(safeTiles < 2)
  {
    if(tile == goal) {return goal;}
    local slope = AITile.GetSlope(tile);  
    
    if(AntPathFinder.isBuildable(tile, ignoreTiles) && slope == AITile.SLOPE_FLAT)
    {
      safeTiles++;
    }
    else
    {
      safeTiles = 0;
    }
    
    if(display_thinking == 1) { AISign.BuildSign(tile, "{-}");}
    tile = CrowPathFinder.GetNextClosestTileAny(tile);
  }
  if(display_thinking == 1) { AISign.BuildSign(tile, "{*}");}
  return tile;
}


function CrowPathFinder::GetNextClosestTile(tile)
{
  local nodelist = AntHill.FindCompassPointsNunitsAway(tile, 1);
  
  nodelist.Valuate(this.IsInPathValuator, path);
  nodelist.KeepValue(0);
   
  nodelist.Valuate(DistanceValuator, goal);
  nodelist.Sort(AIAbstractList.SORT_BY_VALUE, true);
  
  local maxCount = nodelist.Count();
  local curCount = 0;
  local node = nodelist.Begin();
  while(!this.SlopeDirectionValuator(node, last) && curCount <= maxCount)
  {
    curCount++;
    node = nodelist.Next();
  }
  
  if(curCount >= maxCount)
  {
    return null;
  }
  
  return node;
}

function CrowPathFinder::GetNextClosestTileAny(tile)
{
  local nodelist = AntHill.FindCompassPointsNunitsAway(tile, 1);
  
   
  nodelist.Valuate(DistanceValuator, goal);
  nodelist.Sort(AIAbstractList.SORT_BY_VALUE, true);
  
  local node = nodelist.Begin();
  return node;
}

function CrowPathFinder::IsInPathValuator(idx, path)
{
  foreach(i, v in path)
  {
    if(idx == i) {return true;}
  }
  return false;
}

function CrowPathFinder::SlopeDirectionValuator(tile, last)
{
 if(last == null) {return true;}
 local slopeCurrent = AITile.GetSlope(current);
 local slopeTile = AITile.GetSlope(tile);
 local slopeLast = AITile.GetSlope(last);
 if(slopeCurrent == AITile.SLOPE_FLAT && slopeTile == AITile.SLOPE_FLAT && slopeLast == AITile.SLOPE_FLAT) {return true;}
 
 local cardinal = CrowPathFinder.DetermineDirection(tile, last);

 return cardinal;
}


function CrowPathFinder::DistanceValuator(tile, goal)
{
 return AITile.GetDistanceSquareToTile(tile, goal);
}