class StationManager
{
  offset = 5;
  
  industryIndex = null;
  Location = null;
  Length = null;
  DepotFront = null;
  Direction = null;
  StationId = null;
  industryLocation = null;
  Tiles = null;
  JoinTiles = null;
  name = null;
  Depot = null;
  NumPlatforms = 1;
  TrackSignals = null;
  testMode = false;
  
  OutOfCash = false;
  
  constructor(industryIndex, name, NumPlatforms = 2)
  {
    this.industryIndex = industryIndex;
    this.name = name;
    this.industryLocation = AIIndustry.GetLocation(industryIndex);
    this.Location = null;
    this.DepotFront = null;
    this.Direction = null;
    this.Length = null;
    Tiles = AIList();
    JoinTiles = AIList();
    Depot = null;
    this.NumPlatforms = NumPlatforms;
    StationId = null;
    TrackSignals = [];
    testMode = false;
  }
}


function StationManager::PlaceProducingStationTest(length)
{
  testMode = true;
  return this.PlaceProducingStation(length);
  testMode = false;
}

function StationManager::PlaceProducingStation(length)
{
  local producingTiles = AITileList_IndustryProducing(industryIndex, 2);
  local result = PlaceStation(length, producingTiles);
  if(result == true && testMode == false)
  {
    this.PlaceDepot();
  }
  Tiles.Valuate(Inverse);
  return result;
}


function StationManager::Inverse(integer)
{
  return integer * -1;
}


function StationManager::PlaceConsumingStationTest(length, producer)
{
  testMode = true;
  return this.PlaceConsumingStation(length, producer);
  testMode = false;
}

function StationManager::PlaceConsumingStation(length, producer)
{
  local acceptingTiles = AITileList_IndustryAccepting(industryIndex, 2);
  acceptingTiles.Valuate(AITile.GetDistanceManhattanToTile, producer);
  acceptingTiles.Sort(AIAbstractList.SORT_BY_VALUE, true); 
  return PlaceStation(length, acceptingTiles);
}

function StationManager::PlaceStation(length, tileList)
{  
  local result = TryLocations(length, tileList);
  
  foreach(tile, v in Tiles)
  {
    //add border tiles to list of tiles for joining track to.
    if((v + 1) % (length + offset) == 0 || v == 1 || (v - 1) % (length + offset) == 0)
    {
      JoinTiles.AddItem(tile, 0);
    }
  }
  
  
  if(testMode == true)
  {
    return result;
  }
  
  if(result == true)
  {
    result = _PlaceStation(length); 
  }
  
  return result;
}

function StationManager::CreateFeatures(length)
{
  foreach(tile, v in Tiles)
  {
    BringTileToLevel(tile);
  }
  
  Tiles.Sort(AIAbstractList.SORT_BY_VALUE, true);
  foreach(tile, v in Tiles)
  {
    //build parrallel rails.
    if(v % (length + offset) != 0 && (v + 1) % (length + offset) != 0)
    {
      AIRail.BuildRailTrack(tile, Direction);
    }
    
    //make ends into "diamonds"
    if((v + 1) % (length + offset) == 0 || v == 1 || (v - 1) % (length + offset) == 0)
    {
      this.BuildAllTrack(tile);
    }
  }
  //AIRail.BuildRailTrack(DepotFront, Direction);
  
  
  foreach(i, v in TrackSignals)
  {
    AIRail.BuildSignal(v.tile, v.front, AIRail.SIGNALTYPE_NORMAL);
  }
  
  //AISign.BuildSign(Depot, "DEPOT");
  //AISign.BuildSign(DepotFront, "BACK");
  //AISign.BuildSign(Location, "L");

  
}


function StationManager::BuildAllTrack(tile)
{
      
      AIRail.BuildRailTrack(tile, AIRail.RAILTRACK_NW_NE);
      AIRail.BuildRailTrack(tile, AIRail.RAILTRACK_SW_SE);
      AIRail.BuildRailTrack(tile, AIRail.RAILTRACK_NW_SW); 
      AIRail.BuildRailTrack(tile, AIRail.RAILTRACK_NE_SE);
      AIRail.BuildRailTrack(tile, AIRail.RAILTRACK_NE_SW);
      AIRail.BuildRailTrack(tile, AIRail.RAILTRACK_NW_SE);
}

function StationManager::GetStationRectangle(tile, length, direction, platforms)
{
  assert(tile != null);
  platforms++;
  
  local tiles = AIList();
  local dx = 0;
  local dy = 0;
  if(direction == AIRail.RAILTRACK_NW_SE)
  {
    dx = 0;
    dy = 1;
  }
  
  if(direction == AIRail.RAILTRACK_NE_SW)
  {
    dx = 1;
    dy = 0;
  }

  for(local i = 0; i < platforms; i++)
  {
    //print("dy * i == " + (dy * i) + ", dx * i ==" + (dx * i));
    local newtile = tile + AIMap.GetTileIndex(dy * i, dx * i);
    tiles.AddList(GetTiles(newtile, dx, dy, length, i * length ));
  }

  return tiles;

}

function StationManager::GetTiles(tile, dx, dy, length, startIndex)
{
  local tiles = AIList();
  local count = startIndex;
  for(local i = 0; i < length; i++)
  {
    tiles.AddItem(tile, ++count);
    tile = tile + AIMap.GetTileIndex(dx, dy); 
  }
  return tiles;
}



function StationManager::IndustryDistanceValuator(tile, location)
{
  local score = AITile.GetDistanceManhattanToTile(tile, location);
  return score;
}

function StationManager::TryLocations(length, tileList)
{

  local directions = [AIRail.RAILTRACK_NE_SW, AIRail.RAILTRACK_NW_SE];
  local result = false;
  
  tileList.Valuate(IndustryDistanceValuator, industryLocation);
  tileList.Sort(AIAbstractList.SORT_BY_VALUE, false);
  
  //foreach(tile, v in tileList)
  //{
  //  AISign.BuildSign(tile, v + "");
  //}
  
  foreach(tile, v in tileList)
    {
    if(AIStation.GetStationID(tile) != AIStation.STATION_INVALID) 
    {
      AILog.Warning("Already a station here."); 
      continue;
    }
    
    foreach(i, direction in directions)
    {
      local localTestMode = AITestMode();
      result = AIRail.BuildRailStation( tile, direction, NumPlatforms + 1, length + offset, AIStation.STATION_NEW ); 
      if(AIError.GetLastErrorString() == "ERR_LOCAL_AUTHORITY_REFUSES") 
      {
        AILog.Warning("Local Authority Refuses");
        this.PlantTrees(5);
      }
      localTestMode = null;
      if(result == true)
      {
        this.Direction = direction;
        Tiles = GetStationRectangle(tile, length + offset, Direction, NumPlatforms);
        this.GetStationFeatures(Tiles, length);
        return true;
      }  
    }  
  }
 return false;
}


function StationManager::_PlaceStation(length)
{
  assert(this.Location != null);
  local tile = this.Location;
  local result = AIRail.BuildRailStation( Location, Direction, NumPlatforms, length, AIStation.STATION_NEW );
  
  if(result == true)
  {
    StationId = AIStation.GetStationID(Location);
    AIStation.SetName(StationId, this.name);
  }
  
  this.CreateFeatures(length);
  
  //foreach(tile, i in Tiles)
  //{
  //  AISign.BuildSign(tile, i + "");
  //}
  
  return result;
}

function StationManager::LevelTiles(radius)
{
  print("Leveling Around Industry with radius == " + radius);
  if(radius == 0) {return;}
  
  local tile1 = industryLocation + AIMap.GetTileIndex(-radius, - radius);
  local tile2 = industryLocation + AIMap.GetTileIndex(radius + 3, radius + 3);
  local tile3 = industryLocation + AIMap.GetTileIndex(-radius, radius + 3);
  local tile4 = industryLocation + AIMap.GetTileIndex(radius + 3, -radius);
  
  AITile.LevelTiles(industryLocation, tile1);
  AITile.LevelTiles(industryLocation, tile2);
  AITile.LevelTiles(industryLocation, tile3);
  AITile.LevelTiles(industryLocation, tile4);
  
  this.PlantTrees(20);
}


function StationManager::GetTownRating()
{
  local town = AITile.GetClosestTown(industryLocation);
  local rating = AITown.GetRating(town, AICompany.COMPANY_SELF );
  return rating;
}

function StationManager::PlantTrees(radius)
{
  local town = AITile.GetClosestTown(industryLocation);
  local name = AITown.GetName(town);
  local rating = AITown.GetRating(town, AICompany.COMPANY_SELF );
  local townLocation = AITown.GetLocation(town);
  print(name + " town rating is  " + rating +"."); 

  local tile1 = townLocation + AIMap.GetTileIndex(-radius, - radius);
  local count = 0;
  while(rating  == AITown.TOWN_RATING_APPALLING ||  rating  == AITown.TOWN_RATING_VERY_POOR || rating == AITown.TOWN_RATING_POOR )
  {
    rating = AITown.GetRating(town, AICompany.COMPANY_SELF );
    print (name + ": Planting trees and waiting to improve rating "+ rating +".");
    AITile.PlantTreeRectangle (tile1, radius, radius);
    AITile.PlantTreeRectangle (townLocation, radius, radius);
    if(AICompany.GetBankBalance(AICompany.COMPANY_SELF) > 1000000)
    {
      //burn money to increase rating.
      AIController.Sleep(100);
    }
    else
    {
      //count on time to increase rating more than tree planting.
      if(count++ > 5) {break;}
      AIController.Sleep(2000);
    }
  }
}


function StationManager::GetRadiusTiles(location, radius)
{
  local tiles = [];  
  tiles.push(location + AIMap.GetTileIndex(0, radius));
  tiles.push(location + AIMap.GetTileIndex(radius, 0));
  tiles.push(location + AIMap.GetTileIndex(0, -radius));
  tiles.push(location + AIMap.GetTileIndex(-radius, 0));
  return tiles;
}

function StationManager::PlaceDepot()
{
  
  print("PlaceDepot() Depot:"+ this.Depot + " DepotFront:" + this.DepotFront);
  local depotResult = AIRail.BuildRailDepot(this.Depot, this.DepotFront);
  return depotResult;
}


function StationManager::GetStationFeatures(tileList, length)
{
  print ("GetStationFeatures()");
  
  Length = length;
  local temp = AIList();
  Tiles.AddList(tileList);
  
  //Mark Depot
  temp.Clear();
  temp.AddList(tileList);
  temp.KeepValue( ((length + offset) * NumPlatforms) - 1 );
  DepotFront = temp.Begin();
  
  temp.Clear();
  temp.AddList(tileList);
  temp.KeepValue((length + offset) * NumPlatforms); 
  Depot =  temp.Begin();
    
   //Mark entry and exit signals.
  temp.Clear();
  temp.AddList(tileList);
  temp.KeepValue(3); 
  this.Location = temp.Begin();
  
  //Mark Depot
  temp.Clear();
  temp.AddList(tileList);
  temp.KeepValue( ((length + offset) * (NumPlatforms + 1)) - 1 );
  local DepotSideSignalFront = temp.Begin();
  
  temp.Clear();
  temp.AddList(tileList);
  temp.KeepValue( ((length + offset) * (NumPlatforms + 1)) - 2 );
  local DepotSideSignal = temp.Begin();
  
  TrackSignals.push({tile = DepotSideSignal, front = DepotSideSignalFront});
  
  temp.Clear();
  temp.AddList(tileList);
  temp.KeepValue( ((length + offset) * NumPlatforms) + 2 );
  local frontSignal = temp.Begin();
  
  temp.Clear();
  temp.AddList(tileList);
  temp.KeepValue( ((length + offset) * NumPlatforms) + 3 );
  local frontSignalFront = temp.Begin(); 
  TrackSignals.push({tile = frontSignal, front = frontSignalFront});
  
  
  //Mark side signals;
  local front = null;
  local tile = null;
  local value = 0;
  for(local i = 0; i < NumPlatforms; i++)
  {
    for(local j = 0; j < (length + offset); j++)
    {
      value++;
      if(j == 0 || j == (length + offset) -4)
      {
        temp.Clear();
        temp.AddList(tileList);
        temp.KeepValue( value );
        front = temp.Begin();      
      }
      
      
      if(j == 1  || j == (length + offset) -3)
      {
        temp.Clear();
        temp.AddList(tileList);
        temp.KeepValue( value );
        tile = temp.Begin();      
      }
      
      if(tile != null && front != null)
      {
        TrackSignals.push({tile = tile, front = front});
        tile = null;
        front = null;
      }
    }
  }
  
  
  //remove unused tiles.
  for(local i = 1; i <= (length + offset) * (NumPlatforms + 1); i++)
  {
    if(i % (length+offset) == 0)
    {
      Tiles.RemoveValue(i);
    }
  }
  
}

function StationManager::BringTileToLevel(tile)
{
  local slope = AITile.GetSlope(tile);
  if(slope == AITile.SLOPE_FLAT) {return;} 
  local complement = AITile.GetComplementSlope(slope);
  AITile.RaiseTile(tile, complement); 
}

function StationManager::DeleteStation()
{
  foreach(tile, i in Tiles)
  {
    AITile.Demolish(tile);
  }
}