//import("pathfinder.rail", "RailPathFinder", 1);
require("AntPathFinder.nut");
require("AntHill.nut");
require("CrowPathFinder.nut");
require("StationManager.nut");
require("TownManager.nut");


class DustinAI extends AIController
{
  industryInfo = null;
  
  trainsize = 8;
  speedConstant = 0.056;
  PathIterations = null;
  routeCount = 0; 
  cabooseId = null;
  deleteCount = 1;
  stationLen = 5;
  ignoreTiles = null;
  
  
  constructor()
  {
    this.PathIterations = 10;
    this.routeCount = 0;
    this.deleteCount = 1;
    this.cabooseId = FindCaboose();
    ignoreTiles = AIList();
  }
  
  function Start();
 	function Stop();
	function Save();
	function Load(version, data);
}


function DustinAI::Start()
{
  this.Sleep(1);
  print("DustinAI Started.");
  SetCompanyName();
  print("display_thinking = " + AIController.GetSetting("display_thinking"));
  print("bridge_max_len = " + AIController.GetSetting("bridge_max_len")); 
  
  local types = AIRailTypeList();
  AIRail.SetCurrentRailType(types.Begin());
  AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);
  
  AICompany.SetAutoRenewStatus(true);
  
  foreach(town, v in AITownList())
  {
    local tm = TownManager(town);
    local tiles = tm.GetBoundingRect();
    if(tiles != null)
    {
      ignoreTiles.AddList(tm.GetBoundingRect());
    }
  }
  
  
 //Keep running
 while (true)
 {
  local cash = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
  print( "cash == " + cash);
  
  if( cash <= AICompany.GetMaxLoanAmount() * 2)
  {
    AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
  }
  cash = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
  
  if(cash > 200000)
  {
    local result = BuildRoute();
  }
  else
  {
   print("Waiting for more money...");
  }
  
  PayDownLoan();
  
  //Once money becomes less tight, remove the signs from producers that need terraforming to fit a station.
  cash = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
  if(cash > 1000000)
  {
    local signs = AISignList();
    foreach(i, v in signs)
    {
      local name = AISign.GetName(i);
      if(name == "Can't fit a station here."){AISign.RemoveSign(i);}
    }
  }
  
  print("Main Sleep.");
  this.Sleep(2000);
 }
}


function DustinAI::PayDownLoan()
{
  local cash = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
  local interval =  AICompany.GetLoanInterval();
  local loan = AICompany.GetLoanAmount();
  //print(loan + "");

  if (loan == 0) {return;}
  
  local newBalance = loan - interval;
  
  AICompany.SetLoanAmount(newBalance);
  if(AICompany.GetBankBalance(AICompany.COMPANY_SELF) > interval && AICompany.GetLoanAmount() > 0)
  {
    this.Sleep(1);
    PayDownLoan();
  }
}


function DustinAI::Stop()
{
  ClearSigns();
  print(" Stopped");
}


function DustinAI::Save()
{
 local table = {};	
 //print(" Saved");
 return table;
}

function DustinAI::Load(version, data)
{
 print(" Loaded");
}


function DustinAI::SetCompanyName()
{
	if(!AICompany.SetName("Denver & Rio Grande"))
	{
		local i = 2;
		while(!AICompany.SetName("Denver & Rio Grande #" + i))
		{
			i = i + 1;
			if(i > 255)
				break;
		}
	}

	AICompany.SetPresidentName("H.H. Thompson")
}


function DustinAI::GetAmountOfMySigns(industryIndex)
{
  local signs = AISignList();
  
  local signLocations = AIList();
  signLocations.AddList(signs);
  signLocations.Valuate(AISign.GetLocation);
  signLocations.KeepValue(AIIndustry.GetLocation(industryIndex));
  return signLocations.Count();
}

function DustinAI::BuildRoute()
{
  local industryInfo = getIndustryInfo();  
  local industryRank = AIIndustryList();
  industryRank.Valuate(AIIndustry.GetAmountOfStationsAround);
  industryRank.KeepValue(0);
  industryRank.Valuate(GetAmountOfMySigns);
  industryRank.KeepValue(0);
  industryRank.Valuate(DustinAI.ScoreValuator);
  industryRank.Sort(AIAbstractList.SORT_BY_VALUE, false);
  
  local bestIndustry = industryInfo[industryRank.Begin()];
  
  
  
  
  ReplaceSign(bestIndustry.location, "Start " + ++routeCount + " index:" +bestIndustry.index); 
  
  
  local consumers = DustinAI.FindAllConsumersOf(bestIndustry);

  print("Found the distance to each consumer.");
  
  local cashOnHand = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
  print ("Bank Balance:" + cashOnHand); 
  print ("Best Industry Index == " + bestIndustry.index);
  local route = DustinAI.chooseRoute(bestIndustry, consumers);
  if(route == null) {return false;}
  
  //buy trains
  local bestWagon = null;
  local bestEngine = null;
  local trainCost = 0;

  local engineId = null;
  
  foreach(producerCargoIndex, val in bestIndustry.produces)
  {
   bestWagon = DustinAI.FindBestWagon(producerCargoIndex);
   bestEngine = DustinAI.FindBestEngine(bestWagon);
   
   local engineId = AIVehicle.BuildVehicle(route.Depot, bestEngine);
   if(cabooseId != null)
   {
    local caboose = AIVehicle.BuildVehicle(route.Depot, cabooseId);
    AIVehicle.MoveWagon(caboose, 0, engineId, 0);
   }
   
   for(local i = 0; i < this.trainsize; i++)
   {
    local newWagon = AIVehicle.BuildVehicle(route.Depot, bestWagon);    
    AIVehicle.RefitVehicle(newWagon, producerCargoIndex);
    AIVehicle.MoveWagon(newWagon, 0, engineId, 0);
   }

   AIOrder.AppendOrder(engineId, route.Start, AIOrder.AIOF_FULL_LOAD_ANY);
   AIOrder.AppendOrder(engineId, route.End, AIOrder.AIOF_NONE);
   AIVehicle.StartStopVehicle(engineId);
  }    
  
  return true;
}


function DustinAI::ConsumerDistance(idx, producer)
{
    local producerLocation = AIIndustry.GetLocation(producer);
    local distance = AIIndustry.GetDistanceManhattanToTile(idx, producerLocation);
    return distance;
}

function DustinAI::getIndustryInfo()
{
  print("Getting industry info");
  local industryList = AIIndustryList();
  this.industryInfo = {};
  foreach(idx, val in industryList)
  {
    
    local industryType = AIIndustry.GetIndustryType(idx);
    local location = AIIndustry.GetLocation(idx);
    local produces = AIIndustryType.GetProducedCargo(industryType);
    local accepts = AIIndustryType.GetAcceptedCargo(industryType);
    local productionScore = 0;
    foreach(cargoIdx, val in produces)
    {
     local cargoVal = AICargo.GetCargoIncome(cargoIdx, 100, 15);
     local producedLastMonth = AIIndustry.GetLastMonthProduction(idx, cargoIdx);
     //print ("Value:" + cargoVal + " produced:" + producedLastMonth);
     productionScore += cargoVal *  producedLastMonth;
    }
    
    
    
    //insert a table into the array.
    industryInfo[idx] <-
      {
        index = idx,
        type = industryType,
        location = location,
        produces = produces,
        accepts = accepts,
        score = productionScore 
      };
    
  }
 print("Ending industry info collection.");
 return this.industryInfo;

}


function DustinAI::ScoreValuator(industryId)
{
  local industryType = AIIndustry.GetIndustryType(industryId);
  local produces = AIIndustryType.GetProducedCargo(industryType);
  local productionScore = 0;
  foreach(cargoIdx, val in produces)
  {
   local cargoVal = AICargo.GetCargoIncome(cargoIdx, 100, 15);
   local producedLastMonth = AIIndustry.GetLastMonthProduction(industryId, cargoIdx);
   //print ("Value:" + cargoVal + " produced:" + producedLastMonth);
   productionScore += cargoVal *  producedLastMonth;
  }
  return productionScore;
}

function DustinAI::WagonValuator(engineId)
{
  return  AIEngine.GetCapacity(engineId) * AIEngine.GetMaxSpeed(engineId);
}

function DustinAI::FindAllConsumersOf(industry)
{
  local industryIdx = industry.index;
  local cargoIdx = industry.produces.Begin()
  local consumers = AIIndustryList_CargoAccepting(cargoIdx);
  print ("Found " + consumers.Count() + " consumers.");
  consumers.Valuate(DustinAI.ConsumerDistance, industryIdx);
  consumers.Sort(AIAbstractList.SORT_BY_VALUE, true);
  return consumers;
}

function DustinAI::FindBestWagon(cargoIndex)
{   
    local wagons = AIEngineList(AIVehicle.VT_RAIL);
    wagons.Valuate(AIEngine.IsWagon);
    wagons.RemoveValue(0);
    wagons.Valuate(AIEngine.IsValidEngine);
    wagons.RemoveValue(0);
    wagons.Valuate(AIEngine.CanRunOnRail, AIRail.GetCurrentRailType());
    wagons.RemoveValue(0);
    wagons.Valuate(AIEngine.CanRefitCargo, cargoIndex);
    wagons.RemoveValue(0);
    //wagons.AddList(nonRefitWagons);
    wagons.Valuate(DustinAI.WagonValuator);
    if(wagons.Count == 0)
    {
     print ("Warning, no wagons can pull or be refitted to this cargo.");
    }
    return wagons.Begin();
}


function DustinAI::FindCaboose()
{
  local engines = AIEngineList(AIVehicle.VT_RAIL);
  foreach(i, v in engines)
  {
    if(AIEngine.GetName(i) == "Caboose")
    {
      return i;
    }
  }
  return null;
}

function DustinAI::FindBestEngine(wagonId)
{
  local hpPerCar = 175;
  local minHP = 175 * this.trainsize;
  
  local speed = AIEngine.GetMaxSpeed(wagonId);
  if(speed == 0) {speed = 500;}
  local engines = AIEngineList(AIVehicle.VT_RAIL);
  engines.Valuate(AIEngine.IsWagon);
  engines.RemoveValue(1);
  
  engines.Valuate(AIEngine.IsValidEngine);
  engines.RemoveValue(0);
  engines.Valuate(AIEngine.CanRunOnRail, AIRail.GetCurrentRailType());
  engines.RemoveValue(0);
  
  engines.Valuate(AIEngine.GetPower);
  
  engines.Sort(AIAbstractList.SORT_BY_VALUE, false);
  if(engines.GetValue(engines.Begin()) < minHP ) //no engine can pull the wagon at it's top speed.
  {
   print("No engine has enough horsepower to pull all the wagons well.");
  }
  else
  {
    engines.RemoveBelowValue(minHP);
  }
  
  
  engines.Valuate(AIEngine.GetMaxSpeed);
  engines.Sort(AIAbstractList.SORT_BY_VALUE, false);
  
  if(engines.GetValue(engines.Begin()) < speed ) //no engine can pull the wagon at it's top speed.
  {
   print("No engine has top speed of wagon. Picking fastest.");
   return engines.Begin();
  }
  
  engines.RemoveBelowValue(speed);
  engines.Valuate(AIEngine.GetPrice);
  engines.Sort(AIAbstractList.SORT_BY_VALUE, true);
  
  print("The cheapest engine to pull '" + AIEngine.GetName(wagonId) + "'' at full speed ("+ speed +") is '" + AIEngine.GetName(engines.Begin()) +"'" );
  return engines.Begin();
  
}


function DustinAI::chooseRoute(bestIndustry, consumers)
{
  local producingTiles = AITileList_IndustryProducing(bestIndustry.index, 3);
  producingTiles = DustinAI.GetBuildableFlatTiles(producingTiles);
  local producingStationManager = StationManager(bestIndustry.index, "Producer " + routeCount);
  local success = false;
  local radius = 1;
  local levelLimit = 6;
  while(success != true && radius <= levelLimit)
  {
    local cash = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
    radius++;
    if(cash > 500000)
    {
      producingStationManager.LevelTiles(radius);
    }
    success = producingStationManager.PlaceProducingStation(stationLen);
  }
  
  if(!success)
  { 
    ReplaceSign(producingStationManager.industryLocation, "Can't fit a station here.");
    return null;
  }
  
  consumers.Sort(AIAbstractList.SORT_BY_VALUE, true)
  
  foreach(idx, val in consumers)
  {
  
    if(AITile.GetDistanceManhattanToTile(AIIndustry.GetLocation(idx), AIIndustry.GetLocation(bestIndustry.index)) < 50) {continue;}
    local route = {};
  
    local consumingTiles = AITileList_IndustryAccepting(idx, 4);
        
    local consumerStationManager = StationManager(idx, "Consumer" + routeCount);
    local success = consumerStationManager.PlaceConsumingStationTest(stationLen, producingStationManager.Location);
    if(!success) {continue;}

    local producerLocation = GetClosestTile(producingStationManager.JoinTiles, consumerStationManager.Location)
    local consumerLocation = GetClosestTile(consumerStationManager.JoinTiles, producingStationManager.Location);
    
    local prepend = AIList();
    local routeStart = producerLocation;
    prepend.AddItem(routeStart, 0);
    
    local postpend = AIList();
    local routeGoal = consumerLocation;    
    postpend.AddItem(routeGoal, 0);
        
    consumingTiles = GetBuildableFlatTiles(consumingTiles);
        
    //local antHill = AntHill(routeStart, routeGoal, producingStationManager.Tiles, postPendTiles);
    local antHill = AntHill(routeStart, routeGoal, AIList(), AIList(), ignoreTiles);
 
    print("Starting Pathfinder.");
    local path = antHill.FindPathMixed();
    print("pathfinder done.");
    ClearPathSigns();
        
    if(path == null) 
    {
      print ("found no path.");
    }
    else
    {
      consumerStationManager._PlaceStation(stationLen);

      print("path len:" + path.Count());
      antHill.BuildRoute(path);
      
      route.Start <- producingStationManager.Location;
      
      route.End <- consumerStationManager.Location;
      route.Depot <- producingStationManager.Depot;
      return route;  
    }
  }
  
  return null;

}

function DustinAI::GetClosestTile(list, goal)
{
  print("Locations to choose from:" + list.Count());
  local tileList = AIList();
  tileList.AddList(list);
  tileList.Valuate(AITile.GetDistanceSquareToTile, goal);
  tileList.Sort(AIAbstractList.SORT_BY_VALUE, true);
  return tileList.Begin();
}

function DustinAI::FindClosestTileFromList(destination, tileList)
{

     local tiles = AIList();
     tiles.AddList(tileList);
     tiles.Valuate(AITile.GetSlope);
     tiles.KeepValue(AITile.SLOPE_FLAT);
     
     tiles.Valuate(DustinAI.CountFlatTiles);
     tiles.KeepAboveValue(3);
     
     tiles.Valuate(AITile.GetDistanceSquareToTile, destination);
     tiles.Sort(AIAbstractList.SORT_BY_VALUE, true);   
     return tiles.Begin();

}

function DustinAI::GetBuildableFlatTiles(tileList)
{
  tileList.Valuate(AITile.IsBuildable)
  tileList.RemoveValue(0);
  tileList.Valuate(AITile.GetSlope);
  tileList.KeepValue(AITile.SLOPE_FLAT);
  return tileList;
}

function DustinAI::CountFlatTiles(tile)
{
  local nodelist = AntHill.FindCompassPointsNunitsAway(tile, 1);
  nodelist.Valuate(AITile.IsBuildable)
  nodelist.RemoveValue(0);
  nodelist.Valuate(AITile.GetSlope);
  nodelist.KeepValue(AITile.SLOPE_FLAT);
  return nodelist.Count();

}


function DustinAI::ClearPathSigns()
{
  local sign = AISignList();
  foreach(i, v in sign)
  {
    local name = AISign.GetName(i);
    if(name == "{.}" || name == "{-}" ||name == "{A}" || name == "{B}" || name == "{+}" || name == "{=}")
    {
      //print("Removing Sign:" + name);
      AISign.RemoveSign(i);
    }
  }

}


function DustinAI::PlaceSignIfNoneAreHere(tile, text)
{
  local signs = AISignList();
  foreach(i, v in signs)
  {
    if(AISign.GetLocation(i) == tile) {return false;}
  }
  
  AISign.BuildSign(tile, text);
  return true;

}

function DustinAI::ReplaceSign(tile, text)
{
  local signs = AISignList();
  foreach(i, v in signs)
  {
    if(AISign.GetLocation(i) == tile) {AISign.RemoveSign(i);}
  }
  
  AISign.BuildSign(tile, text);
}








